from asgiref.sync import sync_to_async

import asyncio
import collections
import json
import logging
import websocket

from django.conf import settings


class LocalWebSocketClient:
    """Wrapper for handy usage of `websocket.WebSocket` within localhost."""
    local_route = ('tasks' if not settings.CHANNEL_TASKS.proxy_route
                   else f'{settings.CHANNEL_TASKS.proxy_route}-local/tasks')
    local_url = f'ws://127.0.0.1:{settings.CHANNEL_TASKS.local_port}/{local_route}/'
    header = {'Content-Type': 'application/json'}

    def __init__(self, **connect_kwargs):
        self.connect_kwargs = connect_kwargs
        self.ws = websocket.WebSocket()
        self.events = collections.defaultdict(list)
        websocket.setdefaulttimeout(self.connect_kwargs.get('timeout', 10))
        self.wsapp = websocket.WebSocketApp(
            self.local_url, header=self.header, on_message=self.on_message, on_error=self.on_message,
        )
        self.expected_events = {}

    def send_locally(self, json_obj):
        self.ws.connect(self.local_url, header=self.header, **self.connect_kwargs)
        self.ws.send(json.dumps(json_obj))
        ws_response = self.ws.recv()
        self.ws.close()

        return ws_response

    def on_message(self, wsapp, message):
        logging.getLogger('django').debug('Received local WS message: %s', message)
        event = json.loads(message)
        self.events[event['content']['status'].lower()].append(event)

        if self.expected_events and self.expected_events_collected:
            wsapp.close()

    def collect_events(self, event_loop):
        return asyncio.wrap_future(asyncio.run_coroutine_threadsafe(
            sync_to_async(self.wsapp.run_forever)(), event_loop))

    @property
    def expected_events_collected(self) -> bool:
        return all(event_type in self.events and len(self.events[event_type]) == count
                   for event_type, count in self.expected_events.items())



import asyncio
import logging
import time

from django import urls
from django.conf import settings

from channels.auth import AuthMiddlewareStack
from channels.routing import ProtocolTypeRouter, URLRouter
from channels.security.websocket import AllowedHostsOriginValidator

from rest_framework import routers

from django_tasks import asgi_setup
from django_tasks.admin_tools import ModelTask, register_task
from django_tasks.consumers import TaskEventsConsumer, TasksRestConsumer
from django_tasks.viewsets import TaskViewSet


@register_task
async def sleep_test(duration, raise_error=False):
    logging.getLogger('django').info('Starting sleep test.')
    await asyncio.sleep(duration)

    if raise_error:
        logging.getLogger('django').info('Sleep test done with raise.')
        raise Exception('Test error')

    logging.getLogger('django').info('Sleep test done with no raise.')
    return f"Slept for {duration} seconds"


@register_task
async def doctask_access_test(instance_ids: list[int]):
    def instance_function(doctask):
        time.sleep(1)
        logging.getLogger('django').info('Retrieved %s', repr(doctask))
        time.sleep(1)

    await ModelTask('django_tasks', 'DocTask', instance_function)(instance_ids)
    await asyncio.sleep(4)


@register_task
async def doctask_deletion_test(instance_ids: list[int]):
    def instance_function(doctask):
        time.sleep(1)
        doctask.delete()
        logging.getLogger('django').info('Deleted %s', repr(doctask))
        time.sleep(1)

    await ModelTask('django_tasks', 'DocTask', instance_function)(instance_ids)


class OptionalSlashRouter(routers.SimpleRouter):
    def __init__(self):
        super().__init__()
        self.trailing_slash = '/?'


http_paths = []


if settings.CHANNEL_TASKS.expose_doctask_api is True:
    drf_router = OptionalSlashRouter()
    drf_router.register('tasks', TaskViewSet)
    http_paths.append(urls.re_path(r'^api/', URLRouter(TasksRestConsumer.get_urls(drf_router))))


http_paths.append(urls.re_path(r'^', asgi_setup.asgi_app))
url_routers = {
    'http': URLRouter(http_paths),
    'websocket': AllowedHostsOriginValidator(AuthMiddlewareStack(
        URLRouter([urls.path('tasks/', TaskEventsConsumer.as_asgi())])
    )),
}
application = ProtocolTypeRouter(url_routers)



from typing import Any

from rest_framework import serializers

from django_tasks import models

from django_tasks.task_inspector import get_coro_info


class DocTaskSerializer(serializers.ModelSerializer):
    registered_task = serializers.SlugRelatedField(
        slug_field='dotted_path', queryset=models.RegisteredTask.objects.all())

    class Meta:
        model = models.DocTask
        read_only_fields = ('id', 'scheduled_at', 'completed_at', 'document')
        fields = ('registered_task', 'inputs', *read_only_fields)

    @classmethod
    def get_task_group_serializer(cls, json_content, *args, **kwargs):
        kwargs.update(dict(many=True, data=json_content))
        many_serializer = cls(*args, **kwargs)
        many_serializer.is_valid(raise_exception=True)

        return many_serializer

    @classmethod
    def create_doctask_group(cls, json_content, *args, **kwargs):
        many_serializer = cls.get_task_group_serializer(json_content, *args, **kwargs)
        doctasks = many_serializer.save()

        return many_serializer, doctasks

    def validate(self, attrs: dict[str, Any]) -> dict[str, Any]:
        self.context['coro_info'] = get_coro_info(attrs['registered_task'].dotted_path, **attrs['inputs'])

        return attrs
